import tempfile
from pathlib import Path

import numpy as np

from pywatemsedem.grasstrips import scale_ktc_with_grass_strip_width, scale_ktc_zhang


def create_ktc(
    composite_landuse,
    cfactor,
    mask,
    ktc_low,
    ktc_high,
    ktc_limit,
    grass=None,
    correction_width=True,
):
    """Create ktc raster based on C-factor raster

    The ktc raster is generated by classifying low and high erosion potential based
    on the C-factor and ktc_limit (i.e. C_factor < ktc_high -> ktc_low, C_factor
    > ktc_limit -> ktc_high).

    The ktc value for landuse rivers, infrastructure and ponds is set to 9999
    (all sediment is routed downwards).

    The ktc values for grass strips are scaled according to their width (see
    :func:`pywatemsedem.ktc.scale_ktc_gdf_grass_strips`).

    Parameters
    ----------
    composite_landuse: numpy.ndarray
        WaTEM/SEDEM composite landuse. (see :ref:`here <watemsedem:prcmap>`).
    cfactor: numpy.ndarray
        C-factor.
    mask: pywatemsedem.geo.rasters.AbstractRaster
        Mask
    ktc_low: float
        Transport coefficient for land covers with low erosion potential
    ktc_high: float
        Transport coefficient for land covers with high erosion potential
    ktc_limit: float
        C-factor to make distinction between ktc_low and ktc_high
    grass: pywatemsedem.geo.vectors.AbstractVector
    correction_width: bool, default True
        Scale kTC according to width of grass strip.

    Returns
    -------
    arr_ktc: numpy.ndarray
        kTC values.
    grass: pywatemsedem.geo.vectors.AbstractVector
        Updated grass strips
    """

    tiff_temp = Path(
        tempfile.NamedTemporaryFile(
            suffix=".tif", prefix="pywatemsedem", delete=False
        ).name
    )
    mask.write(tiff_temp, format="tiff")

    # reclass based on C-factor
    arr_ktc = np.where(cfactor <= ktc_limit, ktc_low, ktc_high)

    # give certain landuse classes an extremely high ktc-value
    arr_ktc = np.where(np.isin(composite_landuse.arr, [-2, -1, -5]), 9999, arr_ktc)

    # set KTC outside modeldomain to zero
    # KTC_arr = np.where(self.catchm.binarr == 1, KTC_arr, 0)

    if correction_width is not None and not grass.is_empty():

        grass._geodata = scale_ktc_gdf_grass_strips(
            grass.geodata,
            ktc_low,
            ktc_high,
        )
        arr_grass = grass.rasterize(
            tiff_temp, composite_landuse.rp.epsg, col="KTC", gdal=True
        )
        arr_ktc = np.where(
            np.logical_and(composite_landuse.arr == -6, arr_grass != mask.rp.nodata),
            arr_grass,
            arr_ktc,
        )
        arr_ktc = np.where(
            np.logical_and(composite_landuse.arr == -6, arr_grass == mask.rp.nodata),
            ktc_low,
            arr_ktc,
        )
    else:
        arr_ktc = np.where(composite_landuse.arr == -6, ktc_low, arr_ktc)

    return arr_ktc, grass


def scale_ktc_gdf_grass_strips(gdf_grass_strips, ktc_low, ktc_high):
    """Scale the ktc values for grass strips in the grass strips dataframe

    Parameters
    ----------
    gdf_grass_strips: geopandas.GeoDataFrame
        Grass strips geodataframe, 'width' and 'scale_ktc' should be present as
        columns.
    ktc_low: float
        Transport coefficient for land covers with low erosion potential
    ktc_high: float
        Transport coefficient for land covers with high erosion potential

    Returns
    -------
    gdf_grass_strips: geopandas.GeoDataFrame
        Updated with the KTC value and the linear sediment trapping efficiency
        (STE_linear).
        For definition STE_linear,
        :func:`pywatemsedem.grasstrips.scale_ktc_with_grass_strip_width`.

    Notes
    -----
    Whether the kTC is scaled for the grass strips depends on the grass strip
    attribute "scale_ktc". If it is not scaled, the ktc_low-value is assigned.
    """
    if ("width" not in gdf_grass_strips.columns) or (
        "scale_ktc" not in gdf_grass_strips.columns
    ):
        msg = "Columns 'BREEDTE' and 'scale_ktc' should be in grass strips dataframe"
        return KeyError(msg)

    arr_ktc_grass, arr_ste = scale_ktc_with_grass_strip_width(
        gdf_grass_strips["width"].values,
        scale_ktc_zhang,
        ktc_high=ktc_high,
    )

    try:
        arr_scale_mask = np.array(gdf_grass_strips["scale_ktc"], dtype=bool)
    except TypeError:
        msg = (
            "Type of 'scale_ktc' column cannot be converted to boolean. Cannot "
            "mask 'KTC' column."
        )
        raise TypeError(msg)

    arr_ktc_grass[~arr_scale_mask] = ktc_low
    arr_ste[~arr_scale_mask] = np.nan

    gdf_grass_strips["KTC"] = arr_ktc_grass
    gdf_grass_strips["STE_linear"] = arr_ste

    return gdf_grass_strips



import shutil
from pathlib import Path

import pytest
from conftest import ini_file

from pywatemsedem.io.ini import add_field, get_item_from_ini, modify_field


class TestModifyField:
    """Class combining all tests for the modify_field function"""

    def test_modify(self, tmp_path):
        """Test basic func"""
        modify_field(ini_file, "User Choices", "max kernel", 3)

    def test_wrong_section(self, tmp_path):
        """Expect error when the section does not exist"""
        with pytest.raises(KeyError) as excinfo:
            modify_field(ini_file, "User Choice", "max kernel", 3)
        assert "Section 'User Choice' does not exist" in str(excinfo.value)

    def test_wrong_key(self, tmp_path):
        """Expect error when the to-modify-key does not exist"""
        with pytest.raises(KeyError) as excinfo:
            modify_field(ini_file, "User Choices", "max kernelzz", 3)
        assert "Key 'max kernelzz' does not exist" in str(excinfo.value)


class TestAddField:
    """Class combining all tests for the add field section"""

    def test_add(self, tmp_path):
        """Test basic func"""
        shutil.copy(ini_file, tmp_path / "test.ini")
        add_field(tmp_path / "test.ini", "User Choices", "max kernelzzz", 3)

    def test_wrong_section(self, tmp_path):
        """Expect error when the section does not exist"""
        with pytest.raises(KeyError) as excinfo:
            add_field(ini_file, "User Choice", "max kernel", 3)
        assert "Section 'User Choice' does not exist" in str(excinfo.value)

    def test_key_exists(self, tmp_path):
        """Expect error when the to-add-key already exists"""
        with pytest.raises(KeyError) as excinfo:
            add_field(ini_file, "User Choices", "max kernel", 3)
        assert "Key 'max kernel' already exist in" in str(excinfo.value)


class TestGetItemFromIni:
    """Class combining all tests for the get_item_from_ini function"""

    def test_get_string_from_ini(self):
        """Test if a string can be retrieved from an ini-file"""
        value = get_item_from_ini(
            ini_file, section="Test Section", option="string_option", dtype=str
        )
        expected_string = "This is a string"
        assert value == expected_string

    def test_get_float_from_ini(self):
        """Test if a float can be retrieved from an ini-file"""
        value = get_item_from_ini(
            ini_file, section="Test Section", option="float_option", dtype=float
        )
        expected_float = 0.03
        assert value == expected_float

    def test_get_int_from_ini(self):
        """Test if an int can be retrieved from an ini-file"""
        value = get_item_from_ini(
            ini_file, section="Test Section", option="int_option", dtype=int
        )
        expected_int = 10
        assert value == expected_int

    def test_get_bool_from_ini(self):
        """Test if a bool can be retrieved from an ini-file"""
        value = get_item_from_ini(
            ini_file, section="Test Section", option="bool_option", dtype=bool
        )
        expected_bool = True
        assert value == expected_bool

    def test_non_existant_section(self):
        """Test if a ValueError is raised when a value from a non-existing section in the ini-file is wanted"""
        with pytest.raises(ValueError) as excinfo:
            get_item_from_ini(
                ini_file,
                section="Not existing section",
                option="bool_option",
                dtype=bool,
            )
        assert ("Section Not existing section does not exist in ini-file") in str(
            excinfo.value
        )

    def test_non_existant_option(self):
        """Test if a ValueError is raised when a value from a non-existing option in the ini-file is wanted"""
        with pytest.raises(ValueError) as excinfo:
            get_item_from_ini(
                ini_file,
                section="Test Section",
                option="Not existing option",
                dtype=bool,
            )
        assert (
            "Option Not existing option does not exist in ini-file (section Test Section)"
        ) in str(excinfo.value)

    def test_wrong_dtype_from_ini(self):
        """Test if a TypeError is raised
        when no correct dtype is given"""
        with pytest.raises(TypeError) as excinfo:
            get_item_from_ini(
                ini_file,
                section="Test Section",
                option="bool_option",
                dtype="this is not a dtype",
            )
        assert ("not a correct Type") in str(excinfo.value)

    def test_wrong_ini_file(self):
        """Test if a FileNotFoundError is raised when no correct ini file is given"""
        with pytest.raises(FileNotFoundError) as excinfo:
            get_item_from_ini(
                Path("This/File/Does/Not/Exist.ini"),
                "Test Section",
                "bool_option",
                bool,
            )
        assert "This/File/Does/Not/Exist.ini does not exist" in str(excinfo.value)
